/**
 * Lynx.cpp - Lynxmotion 5 Robotic Arm Interface
 * CSE 490I Neurobotics Winter 2007
 * Josh Storz <jstorz@cs.washington.edu>
 */

#include "Lynx.h"

const double Lynx::DEGS_PER_PULSE = 180.0 / (Lynx::PULSE_MAX - Lynx::PULSE_MIN);
const double Lynx::RADS_PER_PULSE =  M_PI / (Lynx::PULSE_MAX - Lynx::PULSE_MIN);

Lynx::Lynx()
{
  // initialize with default settings
  hcomm = NULL;
  comm_port = 1;
  comm_baud = 115200;
}

Lynx::~Lynx()
{
  // free comm port
  if (isConnected())
    close();
  if (hcomm != NULL)
    CloseHandle(hcomm);
}

bool Lynx::open()
{
  if (isConnected())
    return false;

  // open comm port
  
  char config_port[10];
  sprintf_s(config_port, 10, "\\\\.\\COM%d", comm_port);

  hcomm = CreateFile(config_port, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, 0, NULL);

  if (hcomm == INVALID_HANDLE_VALUE)
  {
    hcomm = NULL;
    return false;
  }

  // configure communications port
  //   no parity, 8 bits, 1 stop bit, no flow control

  DCB dcb;
  memset(&dcb, 0, sizeof(dcb));
  dcb.DCBlength = sizeof(dcb);
  GetCommState(hcomm, &dcb);

  char config_str[50];
  sprintf_s(config_str, 50, "baud=%d parity=N data=8 stop=1", comm_baud);

  if (!SetupComm(hcomm, 128, 128) || !BuildCommDCB(config_str, &dcb))
  {
    CloseHandle(hcomm);
    hcomm = NULL;
    return false;
  }

  dcb.fOutxCtsFlow = FALSE;
  dcb.fOutxDsrFlow = FALSE;
  dcb.fDtrControl = DTR_CONTROL_DISABLE;
  dcb.fDtrControl = FALSE;
  dcb.fDsrSensitivity = FALSE;
  dcb.fOutX = FALSE;
  dcb.fInX = FALSE;
  dcb.fNull = FALSE;
  dcb.fRtsControl = RTS_CONTROL_DISABLE;
  dcb.fAbortOnError = FALSE;

  if (!SetCommState(hcomm ,&dcb))
  {
    CloseHandle(hcomm);
    hcomm = NULL;
    return false;
  }

  // set initial servo positions (same start values)
  int num = 5;
  int ch[] = {0, 1, 2, 3, 4};
  int pw[] = {1500, 1200, 1500, 1800, 1500};
  int spd[] = {0, 0, 0, 0, 0};
  if (!sendMoveCmd(num, ch, pw, spd))
  {
    close();
    return false;
  }

  // wait for servos to move
  Sleep(200);
  
  // retrieve initial servo position (should be the same as set position above ^^^)
  if (!updatePos())
  {
    close();
    return false;
  }

  return true;
}

bool Lynx::close()
{
  if (!isConnected())
    return false;

  CloseHandle(hcomm);
  hcomm = NULL;

  return true;
}

void Lynx::setComm(int port)
{
  if (isConnected())
    return;
  
  comm_port = port;
}

int Lynx::getComm()
{
  return comm_port;
}

void Lynx::setBaud(int rate)
{
  if (isConnected())
    return;

  comm_baud = rate;
}

int Lynx::getBaud()
{
  return comm_baud;
}

bool Lynx::setCmd(Lynx::LynxCmd cmd)
{
  if (!isConnected())
    return false;

  int num = 0;
  int ch[5];
  int pw[5];
  int spd[5];

  updatePos();
  
  // merge with current command
  for (int i = 0; i < 5; i++)
  {
    if (cmd.joint_set[i])
    {
      if (cmd.joint_spd[i] == 0)
      {
        // stop joint movement
        ch[num] = i;
        pw[num] = pw_cur[i];
        spd[num] = 0;
        cmd_cur.joint_set[i] = false;
        num++;
      }
      else if (cmd.joint_spd[i] < 0)
      {
        // as fast as possible
        ch[num] = i;
        pw[num] = radsToPulse(cmd.joint_tgt[i]);
        spd[num] = -1;
        cmd_cur.joint_set[i] = true;
        cmd_cur.joint_tgt[i] = cmd.joint_tgt[i];
        cmd_cur.joint_spd[i] = cmd.joint_spd[i];
        num++;
      }
      else
      {
        // at specified speed
        ch[num] = i;
        pw[num] = radsToPulse(cmd.joint_tgt[i]);
        spd[num] = radsToPulse(cmd.joint_spd[i]) - PULSE_MIN;
        cmd_cur.joint_set[i] = true;
        cmd_cur.joint_tgt[i] = cmd.joint_tgt[i];
        cmd_cur.joint_spd[i] = cmd.joint_spd[i];
        num++;
      }
    }
  }

  return sendMoveCmd(num, ch, pw, spd);
}

bool Lynx::sendMoveCmd(int num, int *ch, int *pw, int *spd)
{
  char buf[100];
  char cmdbuf[20];
  strcpy_s(buf, 100, "");

  // build movement command based on input
  for (int i = 0; i < num; i++)
  {
    if (spd[i] > 0)
      sprintf_s(cmdbuf, 20, "#%d P%d S%d ", ch[i], pw[i], spd[i]);
    else
      sprintf_s(cmdbuf, 20, "#%d P%d ", ch[i], pw[i], spd[i]);
    strcat_s(buf, 100, cmdbuf);
  }
  strcat_s(buf, 100, "\r");

  // send to controller
  DWORD written = 0;
  if (!WriteFile(hcomm, buf, (DWORD)strlen(buf), &written, NULL))
  {
    close();
    return false;
  }

  return true;
}

bool Lynx::updatePos()
{
  char readcmd[] = "QP 0 QP 1 QP 2 QP 3 QP 4\r";
  unsigned char buf[5];

  // request pulse widths
  DWORD written = 0;
  if (!WriteFile(hcomm, readcmd, (DWORD)strlen(readcmd), &written, NULL))
  {
    close();
    return false;
  }

  // read response
  DWORD read = 0;
  if (!ReadFile(hcomm, buf, 5, &read, NULL))
  {
    close();
    return false;
  }

  // update state
  for (int i = 0; i < 5; i++)
  {
    pw_cur[i] = (int)buf[i] * 10;
    pos_cur.joint_pos[i] = pulseToRads(pw_cur[i]);
  }

  return true;
}
